@@ -3,6 +3,8 @@
 Namespace for complex SDK functional.
 """
 
+from ambra_sdk.addon.dicom import Dicom
+from ambra_sdk.addon.job import Job
 from ambra_sdk.addon.study import Study
 
 
@@ -20,3 +22,5 @@ class Addon:
     def _init_namespaces(self):
         """Init addon namespaces."""
         self.Study = Study(self._api)
+        self.Job = Job(self._api)
+        self.Dicom = Dicom(self._api)
