@@ -954,3 +954,47 @@ class SfdcJwtMissingFields(PreconditionFailed):
 
 class SfdcJwtNotHash(PreconditionFailed):
     """SfdcJwtNotHash."""
+
+
+class RoleNamespaceMismatch(PreconditionFailed):
+    """RoleNamespaceMismatch."""
+
+
+class BrandNotAllowed(PreconditionFailed):
+    """BrandNotAllowed."""
+
+
+class NoDestination(PreconditionFailed):
+    """NoDestination."""
+
+
+class NotMatched(PreconditionFailed):
+    """NotMatched."""
+
+
+class RequestClosed(PreconditionFailed):
+    """RequestClosed."""
+
+
+class InvalidDistributedDestination(PreconditionFailed):
+    """InvalidDistributedDestination."""
+
+
+class InvalidWebhookSetup(PreconditionFailed):
+    """InvalidWebhookSetup."""
+
+
+class InvalidConfig(PreconditionFailed):
+    """InvalidConfig."""
+
+
+class SiteNotAllowed(PreconditionFailed):
+    """SiteNotAllowed."""
+
+
+class InvalidSignature(PreconditionFailed):
+    """InvalidSignature."""
+
+
+class OnlyOne(PreconditionFailed):
+    """OnlyOne."""
