@@ -1,13 +1,13 @@
 """Storage study namespace."""
 
 import json
-from io import BufferedReader
 from typing import Any, Dict, Optional, Set, Union
 
 from box import Box, BoxList
 from requests import Response
 
 from ambra_sdk.exceptions.storage import (
+    Hl7ReportDataNotConvertedToDicom,
     InconsistencyConflict,
     NotFound,
     PermissionDenied,
@@ -17,6 +17,7 @@ from ambra_sdk.exceptions.storage import (
 )
 from ambra_sdk.storage.bool_to_int import bool_to_int
 from ambra_sdk.storage.request import PreparedRequest, StorageMethod
+from ambra_sdk.types import RequestsFileType
 
 
 def validate_size_x_string(size_str: str):
@@ -239,6 +240,7 @@ class Study:
         namespace: str,
         study_uid: str,
         image_uid: str,
+        x_ambrahealth_job_id: Optional[str] = None,
         only_prepare: bool = False,
     ) -> Union[Response, PreparedRequest]:
         """Deletes a single image from a study.
@@ -255,6 +257,7 @@ class Study:
         :param namespace: Namespace (Required).
         :param study_uid: Study uid (Required).
         :param image_uid: Image uid (Required).
+        :param x_ambrahealth_job_id: X-AmbraHealth-Job-Id headers argument
         :param only_prepare: Get prepared request.
 
         :returns: Delete image response
@@ -274,17 +277,18 @@ class Study:
             locals(),
         )
         errors_mapping = {
-            404:
-            NotFound(
-                'Image does not exists.',
-            ),
+            404: NotFound('Image does not exists.'),
         }
+        headers = {}
+        if x_ambrahealth_job_id is not None:
+            headers['X-AmbraHealth-Job-Id'] = x_ambrahealth_job_id
         prepared_request = PreparedRequest(
             storage_=self._storage,
             method=StorageMethod.delete,
             url=url,
             errors_mapping=errors_mapping,
             params=request_data,
+            headers=headers,
         )
         if only_prepare is True:
             return prepared_request
@@ -329,14 +333,8 @@ class Study:
             locals(),
         )
         errors_mapping = {
-            404:
-            NotFound(
-                'Study does not exist.',
-            ),
-            422:
-            UnprocessableEntity(
-                'Request body is empty.',
-            ),
+            404: NotFound('Study does not exist.'),
+            422: UnprocessableEntity('Request body is empty.'),
         }
         prepared_request = PreparedRequest(
             storage_=self._storage,
@@ -1208,13 +1206,14 @@ class Study:
         engine_fqdn: str,
         namespace: str,
         study_uid: str,
-        opened_file: BufferedReader,
+        opened_file: RequestsFileType,
         phi_namespace: Optional[str] = None,
         wrap_images: Optional[bool] = None,
         wrap_html_as_pdf: Optional[bool] = None,
         return_html: Optional[bool] = None,
         synchronous_wrap: Optional[bool] = None,
         static_ids: Optional[bool] = None,
+        x_ambrahealth_job_id: Optional[str] = None,
         only_prepare: bool = False,
     ) -> Union[Response, PreparedRequest]:
         """Posts an attachment to a study.
@@ -1224,7 +1223,10 @@ class Study:
         :param engine_fqdn: Engine FQDN (Required).
         :param namespace: Namespace (Required).
         :param study_uid: Study uid (Required).
-        :param opened_file: Opened file (Required).
+        :param opened_file: Opened file (like in requests) (Required).
+            File object, or may be 2-tuples (filename, fileobj),
+            3-tuples (filename, fileobj, contentype) or
+            4-tuples (filename, fileobj, contentype, custom_headers).
         :param phi_namespace: A string, set to the UUID of
              the namespace where the file was attached if it
              was attached to a shared instance of the study
@@ -1249,6 +1251,7 @@ class Study:
         :param return_html: An integer of value 1 or 0. Return results as
             Content Type text html, instead of
             application json (required for certain browsers)
+        :param x_ambrahealth_job_id: X-AmbraHealth-Job-Id headers argument
         :param only_prepare: Get prepared request.
 
         :returns: posts attachments response
@@ -1300,11 +1303,15 @@ class Study:
             request_arg_names,
             locals(),
         )
+        headers = {}
+        if x_ambrahealth_job_id is not None:
+            headers['X-AmbraHealth-Job-Id'] = x_ambrahealth_job_id
         prepared_request = PreparedRequest(
             storage_=self._storage,
             method=StorageMethod.post,
             url=url,
             params=request_data,
+            headers=headers,
             files=files,
         )
         if only_prepare is True:
@@ -1324,6 +1331,9 @@ class Study:
         include_wrapped_dicoms: Optional[bool] = None,
         stop_on_failure: Optional[bool] = None,
         exclude_viewer: Optional[bool] = None,
+        v3: Optional[bool] = None,
+        roche_directory: Optional[bool] = None,
+        flat_directory: Optional[bool] = None,
         only_prepare: bool = False,
     ) -> Union[Response, PreparedRequest]:
         """Downloads a study ZIP file.
@@ -1350,6 +1360,12 @@ class Study:
         :param exclude_viewer: If "1", viewer app will not be
             included in the "iso", "osx", and "win" bundle.
             The viewer app can be retrieved separately via /download/viewer
+        :param v3: If "1", viewer app will be pro viewer, and format/content
+            will support pro viewer.
+        :param roche_directory: If "1", .dcm files will be organized into
+            StudyDate-ClinicalTrialTimePointID/Modality/SeriesTime-SeriesDescription/ folders (instead of SER000X/ folders).
+        :param flat_directory: If "1", .dcm files will be flatly named
+            IMG0001-IMG{image count}, and not be organized into SER folder.
         :param only_prepare: Get prepared request.
 
         :raises ValueError: Wrong bundle value
@@ -1364,6 +1380,9 @@ class Study:
         )
         stop_on_failure: int = bool_to_int(stop_on_failure)  # type: ignore
         exclude_viewer: int = bool_to_int(exclude_viewer)  # type: ignore
+        v3: int = bool_to_int(v3)  # type: ignore
+        roche_directory: int = bool_to_int(roche_directory)  # type: ignore
+        flat_directory: int = bool_to_int(flat_directory)  # type: ignore
 
         url_template = '/study/{namespace}/{study_uid}/download'
         url_arg_names = {
@@ -1380,6 +1399,9 @@ class Study:
             'include_wrapped_dicoms',
             'stop_on_failure',
             'exclude_viewer',
+            'v3',
+            'roche_directory',
+            'flat_directory',
         }
         url, request_data = self._storage.get_url_and_request(
             url_template,
@@ -1415,6 +1437,7 @@ class Study:
         attachment_id: str,
         hash_arg: str,
         phi_namespace: Optional[str] = None,
+        x_ambrahealth_job_id: Optional[str] = None,
         only_prepare: bool = False,
     ) -> Union[Response, PreparedRequest]:
         """Deletes a study attachment.
@@ -1432,6 +1455,7 @@ class Study:
         :param phi_namespace: A string, set to the UUID of the
             namespace where the file was attached if it was
             attached to a shared instance of the study outside of the original storage namespace
+        :param x_ambrahealth_job_id: X-AmbraHealth-Job-Id headers argument
         :param only_prepare: Get prepared request.
 
         :returns: Delete attachemnt response
@@ -1451,11 +1475,15 @@ class Study:
             request_arg_names,
             locals(),
         )
+        headers = {}
+        if x_ambrahealth_job_id is not None:
+            headers['X-AmbraHealth-Job-Id'] = x_ambrahealth_job_id
         prepared_request = PreparedRequest(
             storage_=self._storage,
             method=StorageMethod.delete,
             url=url,
             params=request_data,
+            headers=headers,
         )
         if only_prepare is True:
             return prepared_request
@@ -1591,6 +1619,7 @@ class Study:
         secondary_study_uid: str,
         delete_secondary_study: Optional[bool] = None,
         series_uids: Optional[str] = None,
+        x_ambrahealth_job_id: Optional[str] = None,
         only_prepare: bool = False,
     ) -> Union[Response, PreparedRequest]:
         """Merge studies.
@@ -1609,6 +1638,7 @@ class Study:
         :param series_uids: A list of one or more comma-separated
             Series Instance UIDs, used to filter images merged
             from secondary study.
+        :param x_ambrahealth_job_id: X-AmbraHealth-Job-Id headers argument
         :param only_prepare: Get prepared request.
 
         :returns: Actually empty response
@@ -1644,12 +1674,16 @@ class Study:
         errors_mapping = {
             500: permission_denied,
         }
+        headers = {}
+        if x_ambrahealth_job_id is not None:
+            headers['X-AmbraHealth-Job-Id'] = x_ambrahealth_job_id
         prepared_request = PreparedRequest(
             storage_=self._storage,
             method=StorageMethod.get,
             url=url,
             errors_mapping=errors_mapping,
             params=request_data,
+            headers=headers,
         )
         if only_prepare is True:
             return prepared_request
@@ -1665,6 +1699,7 @@ class Study:
         new_study_uid: Optional[str] = None,
         keep_image_uids: Optional[str] = None,
         color: Optional[str] = None,
+        x_ambrahealth_job_id: Optional[str] = None,
         only_prepare: bool = False,
     ) -> Union[Response, PreparedRequest]:
         """Produce a new study that is a copy of the old, with specified pixel regions obscured.
@@ -1684,6 +1719,7 @@ class Study:
             of modified copies be same as originals? (default is false)
         :param color: HTML-formatted color (rrggbb) of
             obscured regions (default is black-and-white checkerboard)
+        :param x_ambrahealth_job_id: X-AmbraHealth-Job-Id headers argument
         :param only_prepare: Get prepared request.
 
         :returns: Anonymize study response
@@ -1715,6 +1751,8 @@ class Study:
         headers = {
             'Content-Type': 'application/json; charset=utf-8',
         }
+        if x_ambrahealth_job_id is not None:
+            headers['X-AmbraHealth-Job-Id'] = x_ambrahealth_job_id
         errors_mapping = {
             412:
             PreconditionFailed(
@@ -1855,3 +1893,134 @@ class Study:
         if only_prepare is True:
             return prepared_request
         return prepared_request.execute()
+
+    def hl7_to_sr(
+        self,
+        engine_fqdn: str,
+        namespace: str,
+        study_uid: str,
+        hl7uuid: str,
+        phi_namespace: Optional[str] = None,
+        only_prepare: bool = False,
+    ) -> Union[Response, PreparedRequest]:
+        """Hl7 to sr.
+
+        Gets the HL7 report from services, converts to the DICOM SR
+        and uploads to the storage.
+
+        URL: /study/{namespace}/{studyUid}/hl7/{hl7Uuid}/sr?sid={sid}&phi_namespace={phi_namespace}
+
+        :param engine_fqdn: Engine FQDN (Required).
+        :param namespace: Namespace (Required).
+        :param study_uid: Study uid (Required).
+        :param hl7uuid: hl7 report UUID from services to convert to DICOM SR.
+        :param phi_namespace: A string, set to the UUID of the namespace
+            where the file was attached if it was attached to a shared
+            instance of the study outside of the original storage namespace
+        :param only_prepare: Get prepared request.
+
+
+        :returns: hl7_to_sr result
+        """
+        url_template = '/study/{namespace}/{study_uid}/hl7/{hl7uuid}/sr'
+        url_arg_names = {
+            'engine_fqdn',
+            'namespace',
+            'study_uid',
+            'hl7uuid',
+        }
+        request_arg_names = {
+            'phi_namespace',
+        }
+        url, request_data = self._storage.get_url_and_request(
+            url_template,
+            url_arg_names,
+            request_arg_names,
+            locals(),
+        )
+        errors_mapping = {
+            500:
+            Hl7ReportDataNotConvertedToDicom(
+                'If hl7 report data not converted to DICOM.',
+            ),
+        }
+
+        prepared_request = PreparedRequest(
+            storage_=self._storage,
+            method=StorageMethod.post,
+            url=url,
+            errors_mapping=errors_mapping,
+            params=request_data,
+        )
+        if only_prepare is True:
+            return prepared_request
+        return prepared_request.execute()
+
+    def clone(
+        self,
+        engine_fqdn: str,
+        namespace: str,
+        study_uid: str,
+        phi_namespace: Optional[str] = None,
+        new_image_uids: Optional[bool] = None,
+        new_series_uids: Optional[bool] = None,
+        x_ambrahealth_job_id: Optional[str] = None,
+        only_prepare: bool = False,
+    ) -> Union[Response, PreparedRequest]:
+        """Clone.
+
+        Clones the specified study into new study with new study uid,
+        and generates new series uid and image uids if it's requested.
+
+        URL: /study/{namespace}/{studyUid}/clone?sid={sid}&phi_namespace={phi_namespace}&new_image_uids={true/false}&new_series_uids={true/false}
+
+        :param engine_fqdn: Engine FQDN (Required).
+        :param namespace: Namespace (Required).
+        :param study_uid: Study uid (Required).
+        :param phi_namespace: A string, set to the UUID of the namespace
+            where the file was attached if it was attached to a shared
+            instance of the study outside of the original storage namespace
+        :param new_image_uids: true/false, whether to generate for study new image uids.
+        :param new_series_uids: true/false, whether to generate for study new series uids.
+        :param x_ambrahealth_job_id: X-AmbraHealth-Job-Id headers argument
+        :param only_prepare: Get prepared request.
+
+        :return: new study uid
+        """
+        if new_image_uids is not None:
+            new_image_uids: str = 'true' if new_image_uids else 'false'  # type: ignore
+
+        if new_series_uids is not None:
+            new_series_uids: str = 'true' if new_series_uids else 'false'  # type: ignore
+
+        url_template = '/study/{namespace}/{study_uid}/clone'
+        url_arg_names = {
+            'engine_fqdn',
+            'namespace',
+            'study_uid',
+        }
+        request_arg_names = {
+            'phi_namespace',
+            'new_image_uids',
+            'new_series_uids',
+        }
+        url, request_data = self._storage.get_url_and_request(
+            url_template,
+            url_arg_names,
+            request_arg_names,
+            locals(),
+        )
+        headers = {}
+        if x_ambrahealth_job_id is not None:
+            headers['X-AmbraHealth-Job-Id'] = x_ambrahealth_job_id
+        prepared_request = PreparedRequest(
+            storage_=self._storage,
+            # This method is POST !
+            method=StorageMethod.post,
+            url=url,
+            params=request_data,
+            headers=headers,
+        )
+        if only_prepare is True:
+            return prepared_request
+        return prepared_request.execute()
